import math
import os
import random
import string
import sys

DEFAULT_LENGTH = 32
DEFAULT_SENTENCE = 5
HELP_MESSAGE = '''
===============================================================================
Generates a randomized password.

> passwordy [length] [options]

    length: How many characters. Default %03d.
    options:
        h  : consist entirely of hexadecimal characters.
        b  : consist entirely of binary characters.
        dd : consist entirely of decimal characters.
        default : consist entirely of upper+lower letters.

        p  : allow punctuation in conjunction with above.
        d  : allow digits in conjunction with above.

        l  : convert to lowercase.
        u  : convert to uppercase.
        nd : no duplicates. Each character can only appear once.

Examples:
> passwordy 32 h l
98f17b6016cf08cc00f2aeecc8d8afeb

> passwordy 32 h u
2AA706866BF7A5C18328BF866136A261

> passwordy 32 u
JHEPTKCEFZRFXILMASHNPSTFFNWQHTTN

> passwordy 32 p
Q+:iSKX!Nt)ewUvlE*!+^D}hp+|<wpJ}

> passwordy 32 l p
m*'otz/"!qo?-^wwdu@fasf:|ldkosi`

===============================================================================

Generates a randomized sentence of words.

> passwordy sent [length] [join]

    length : How many words. Default %03d.
    join   : The character that will join words together.
             Default space.

Examples:
> passwordy sent
arrowroot sheared rustproof undo propionic acid

> passwordy sent 8
cipher competition solid angle rigmarole lachrymal social class critter consequently

> passwordy sent 8 _
Kahn_secondary_emission_unskilled_superior_court_straight_ticket_voltameter_hopper_crass

===============================================================================
 '''.strip() % (DEFAULT_LENGTH, DEFAULT_SENTENCE)


def listget(li, index, fallback=None):
    try:
        return li[index]
    except IndexError:
        return fallback

def make_password(length=None, passtype='standard'):
    '''
    Returns a string of length `length` consisting of a random selection
    of uppercase and lowercase letters, as well as punctuation and digits
    if parameters permit
    '''
    if length is None:
        length = DEFAULT_LENGTH

    alphabet = ''

    if 'standard' in passtype:
        alphabet = string.ascii_letters
    elif 'digit_only' in passtype:
        alphabet = string.digits
    elif 'hex' in passtype:
        alphabet = '0123456789abcdef'
    elif 'binary' in passtype:
        alphabet = '01'

    if '+digits' in passtype:
        alphabet += string.digits
    if '+punctuation' in passtype:
        alphabet += string.punctuation
    if '+lowercase' in passtype:
        alphabet = alphabet.lower()
    elif '+uppercase' in passtype:
        alphabet = alphabet.upper()

    alphabet = list(set(alphabet))

    if '+noduplicates' in passtype:
        if len(alphabet) < length:
            message = 'Alphabet "%s" is not long enough to support no-dupe password of length %d'
            message = message % (alphabet, length)
            raise Exception(message)
        password = ''
        for x in range(length):
            random.shuffle(alphabet)
            password += alphabet.pop(0)
    else:
        password = ''.join([random.choice(alphabet) for x in range(length)])
    return password

def make_sentence(length=None, joiner=' '):
    '''
    Returns a string containing `length` words, which come from
    dictionary.common.
    '''
    import dictionary.common as common
    if length is None:
        length = DEFAULT_LENGTH
    words = [random.choice(common.words) for x in range(length)]
    words = [w.replace(' ', joiner) for w in words]
    result = joiner.join(words)
    return result

def urandom_hex(length):
    randbytes = os.urandom(math.ceil(length / 2))
    token = ''.join('{:02x}'.format(x) for x in randbytes)
    token = token[:length]
    return token


def main_password(argv):
    length = listget(argv, 0, DEFAULT_LENGTH)
    options = [a.lower() for a in argv[1:]]

    if '-' in length:
        length = length.replace(' ', '')
        length = [int(x) for x in length.split('-', 1)]
        length = random.randint(*length)

    elif not length.isdigit() and options == []:
        options = [length]
        length = DEFAULT_LENGTH

    length = int(length)

    passtype = 'standard'
    if 'dd' in options:
        passtype = 'digit_only'
    if 'b' in options:
        passtype = 'binary'
    if 'h' in options:
        passtype = 'hex'

    if 'l' in options:
        passtype += '+lowercase'
    elif 'u' in options:
        passtype += '+uppercase'
    if 'p' in options:
        passtype += '+punctuation'
    if 'd' in options:
        passtype += '+digits'
    if 'nd' in options:
        passtype += '+noduplicates'

    return make_password(length, passtype=passtype)

def main_sentence(argv):
    length = listget(argv, 1, DEFAULT_SENTENCE)
    joiner = listget(argv, 2, ' ')

    try:
        length = int(length)
    except ValueError:
        joiner = length
        length = DEFAULT_SENTENCE

    return make_sentence(length, joiner)

def main_urandom(argv):
    length = listget(argv, 1, DEFAULT_LENGTH)
    length = int(length)
    return urandom_hex(length)

def main(argv):
    mode = listget(argv, 0, 'password')
    if 'help' in mode:
        print(HELP_MESSAGE)
        quit()

    if 'sent' in mode:
        print(main_sentence(argv))
    elif 'urandom' in mode:
        print(main_urandom(argv))
    else:
        print(main_password(argv))


if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))



import threading
import time

class Ratelimiter:
    '''
    The Ratelimiter class is used to limit how often you perform some other
    action. Just create a Ratelimiter object with the allowance you need, then
    call `limit()` before doing the thing you wish to ratelimit.

    Example:

        download_limiter = Ratelimiter(allowance=1, period=3)

        for file_url in file_urls:
            download_limiter.limit()
            download(file_url)
    '''
    def __init__(self, allowance, period=1, operation_cost=1, mode='sleep'):
        '''
        allowance:
            Our spending balance per `period` seconds.

        period:
            The number of seconds over which we can perform `allowance` operations.

        operation_cost:
            The default amount to remove from our balance after each operation.
            Pass a `cost` parameter to `self.limit` to use a nondefault value.

        mode:
            'sleep':
                If we do not have the balance for an operation, sleep until we
                do. Then return True every time.

            'reject':
                If we do not have the balance for an operation, do nothing and
                return False. Otherwise subtract the cost and return True.

        Although (allowance=1, period=1) and (allowance=30, period=30) can both
        be described as "once per second", the latter allows for much greater
        burstiness of operation. You could spend the whole allowance in a
        single second, then relax for 29 seconds, for example.
        '''
        if mode not in ('sleep', 'reject'):
            raise ValueError(f'Invalid mode {repr(mode)}.')

        self.allowance = allowance
        self.period = period
        self.operation_cost = operation_cost
        self.mode = mode
        self.lock = threading.Lock()

        self.last_operation = time.time()
        self.balance = 0

    @property
    def gain_rate(self):
        return self.allowance / self.period

    def _limit(self, cost):
        time_diff = time.time() - self.last_operation
        self.balance += time_diff * self.gain_rate
        self.balance = min(self.balance, self.allowance)

        if self.balance >= cost:
            self.balance -= cost
            successful = True

        elif self.mode == 'reject':
            successful = False

        else:
            deficit = cost - self.balance
            time_needed = deficit / self.gain_rate
            time.sleep(time_needed)
            self.balance = 0
            successful = True

        self.last_operation = time.time()
        return successful

    def limit(self, cost=None):
        '''
        See the main class docstring for info about cost and mode behavior.
        '''
        if cost is None:
            cost = self.operation_cost

        with self.lock:
            return self._limit(cost)



def bencode(data):
    '''
    Encode python types to bencode.
    '''
    data_type = type(data)

    encoders = {
        bytes: _encode_bytes,
        int: _encode_int,
        dict: _encode_dict,
        tuple: _encode_list,
        list: _encode_list,
    }

    encoder = encoders.get(data_type, None)
    if encoder is None:
        raise ValueError(f'Invalid data type {data_type}.')
    return encoder(data)

def bdecode(data):
    '''
    Decode bencode to python types.
    '''
    return _decode(data)['result']

# INTERNALS
################################################################################
def _encode_bytes(data):
    '''
    Binary data is encoded as {length}:{bytes}.
    '''
    return b'%d:%s' % (len(data), data)

def _encode_dict(data):
    '''
    Dicts are encoded as d{key}{value}{key}{value}e with the keys in
    lexicographic order.
    Keys must be byte strings
    '''
    result = []
    keys = list(data.keys())
    keys.sort()
    for key in keys:
        result.append(bencode(key))
        result.append(bencode(data[key]))
    result = b''.join(result)
    return b'd%se' % result

def _encode_int(data):
    '''
    Integers are encoded as i{integer}e.
    '''
    return b'i%de' % data

def _encode_list(data):
    '''
    Lists are encoded as l{item}{item}{item}e.
    '''
    result = []
    for item in data:
        result.append(bencode(item))
    result = b''.join(result)
    return b'l%se' % result

def _decode(data):
    if not isinstance(data, bytes):
        raise TypeError(f'bencode data should be bytes, not {type(data)}.')

    identifier = data[0:1]
    if identifier == b'i':
        ret = _decode_int(data)

    elif identifier.isdigit():
        ret = _decode_bytes(data)

    elif identifier == b'l':
        ret = _decode_list(data)

    elif identifier == b'd':
        ret = _decode_dict(data)

    else:
        raise ValueError(f'Invalid initial delimiter "{identifier}".')

    return ret

def _decode_bytes(data):
    colon = data.find(b':')

    start = colon + 1
    size = int(data[:colon])
    end = start + size

    text = data[start:end]
    remainder = data[end:]

    return {'result': text, 'remainder': remainder}

def _decode_dict(data):
    result = {}

    # slice leading d
    remainder = data[1:]

    # Checking [0:1] instead of [0] because [0] returns an int!!!!
    # [0:1] returns b'e' which I want.
    while remainder[0:1] != b'e':
        temp = _decode(remainder)
        key = temp['result']
        remainder = temp['remainder']

        temp = _decode(remainder)
        value = temp['result']
        remainder = temp['remainder']
        result[key] = value

    # slice ending e
    remainder = remainder[1:]
    return {'result': result, 'remainder': remainder}

def _decode_int(data):
    # slide leading i
    data = data[1:]

    end = data.find(b'e')
    if end == -1:
        raise ValueError('Missing end delimiter "e"')
    result = int(data[:end])

    # slice ending e
    remainder = data[end+1:]
    return {'result': result, 'remainder': remainder}

def _decode_list(data):
    result = []

    # slice leading l
    remainder = data[1:]

    while remainder[0:1] != b'e':
        item = _decode(remainder)
        result.append(item['result'])
        remainder = item['remainder']

    # slice ending e
    remainder = remainder[1:]
    return {'result': result, 'remainder': remainder}



import collections
import math
import time

class RateMeter:
    def __init__(self, span):
        '''
        This class is used to calculate a rolling average of
        units per second over `span` seconds.

        Set `span` to None to calculate unit/s over the lifetime of the object
        after the first digest, rather than over a span.
        This saves the effort of tracking timestamps. Don't just use a large number!
        '''
        self.sum = 0
        self.span = span

        self.tracking = collections.deque()
        self.first_digest = None

    def digest(self, value):
        now = time.time()
        self.sum += value

        if self.span is None:
            if self.first_digest is None:
                self.first_digest = now
            return

        earlier = now - self.span
        while len(self.tracking) > 0 and self.tracking[0][0] < earlier:
            (timestamp, pop_value) = self.tracking.popleft()
            self.sum -= pop_value

        if len(self.tracking) == 0 or self.tracking[-1] != now:
            self.tracking.append([now, value])
        else:
            self.tracking[-1][1] += value

    def report(self):
        '''
        Return a tuple containing the running sum, the time span
        over which the rate is being calculated, and the rate in
        units per second.

        (sum, time_interval, rate)
        '''
        # Flush the old values, ensure self.first_digest exists.
        self.digest(0)

        if self.span is None:
            now = math.ceil(time.time())
            time_interval = now - self.first_digest
        else:
            # No risk of IndexError because the digest(0) ensures we have
            # at least one entry.
            time_interval = self.tracking[-1][0] - self.tracking[0][0]

        if time_interval == 0:
            return (self.sum, 0, self.sum)
        rate = self.sum / time_interval
        time_interval = round(time_interval, 3)
        rate = round(rate, 3)
        return (self.sum, time_interval, rate)



import collections
import itertools
import string
import sys

from voussoirkit import basenumber

class Landmark:
    def __init__(self, opener, closer, parser):
        self.opener = opener
        self.closer = closer
        self.parser = parser

def barsplit(chars):
    wordlist = []
    wordbuff = []

    def flush():
        if not wordbuff:
            return
        word = fusk_join(wordbuff)
        wordlist.append(word)
        wordbuff.clear()

    for item in chars:
        if item == '|':
            flush()
        else:
            wordbuff.append(item)
    flush()
    return wordlist

def fusk_join(items):
    form = ''
    fusks = []
    result = []
    for item in items:
        if isinstance(item, str):
            form += item
        else:
            form += '{}'
            fusks.append(item)
    product = itertools.product(*fusks)
    for group in product:
        f = form.format(*group)
        result.append(f)
    return result

def fusk_spinner(items):
    for item in items:
        if isinstance(item, str):
            yield item
        else:
            yield from item

def parse_spinner(characters):
    words = barsplit(characters)
    spinner = fusk_spinner(words)
    return spinner

def fusk_range(lo, hi, padto=0, base=10, lower=False):
    for x in range(lo, hi+1):
        x = basenumber.to_base(x, base)
        x = x.rjust(padto, '0')
        if lower:
            x = x.lower()
        yield x

def parse_range(characters):
    r = ''.join(characters)
    (lo, hi) = r.split('-')
    lo = lo.strip()
    hi = hi.strip()

    lowers = string.digits + string.ascii_lowercase
    uppers = string.digits + string.ascii_uppercase
    lohi = lo + hi
    lower = False
    if all(c in string.digits for c in lohi):
        base = 10
    elif all(c in lowers for c in lohi):
        lower = True
        base = 36
    elif all(c in uppers for c in lohi):
        base = 36
    else:
        base = 62

    if (not lo) or (not hi):
        raise ValueError('Invalid range', r)
    if len(lo) > 1 and lo.startswith('0'):
        padto = len(lo)
        if len(hi) != padto:
            raise ValueError('Inconsistent padding', lo, hi)
    else:
        padto = 0
    lo = basenumber.from_base(lo, base)
    hi = basenumber.from_base(hi, base)

    frange = fusk_range(lo, hi, padto=padto, base=base, lower=lower)
    return frange


landmarks = {
    '{': Landmark('{', '}', parse_spinner),
    '[': Landmark('[', ']', parse_range),
}

def fusker(fstring, landmark=None, depth=0):
    escaped = False
    result = []
    buff = []

    if isinstance(fstring, str):
        fstring = collections.deque(fstring)
    while fstring:
        character = fstring.popleft()
        if escaped:
            buff.append('\\' + character)
            escaped = False
        elif character == '\\':
            escaped = True
        elif landmark and character == landmark.closer:
            buff = [landmark.parser(buff)]
            break
        elif character in landmarks:
            subtotal = fusker(fstring, landmark=landmarks[character])
            buff.extend(subtotal)
        else:
            buff.append(character)
    if not landmark:
        buff = parse_spinner(buff)
    return buff
    return result

if __name__ == '__main__':
    pattern = sys.argv[1]
    fusk = fusker(pattern)
    for result in fusk:
        print(result)



'''
Create the file, or update the last modified timestamp.
'''
import os
import sys

from voussoirkit import clipext
from voussoirkit import winglob

def touch(glob_pattern):
    filenames = winglob.glob(glob_pattern)
    if len(filenames) == 0:
        open(glob_pattern, 'a').close()
        return [glob_pattern]
    else:
        for filename in filenames:
            os.utime(filename)
        return filenames

if __name__ == '__main__':
    glob_patterns = [clipext.resolve(x).strip() for x in sys.argv[1:]]
    for glob_pattern in glob_patterns:
        touch(glob_pattern)



import threading

from voussoirkit import sentinel

PENDING = 'pending'
RUNNING = 'running'
FINISHED = 'finished'
RAISED = 'raised'

NO_RETURN = sentinel.Sentinel('NO_RETURN', truthyness=False)
NO_EXCEPTION = sentinel.Sentinel('NO_EXCEPTION', truthyness=False)

class ThreadPoolException(Exception):
    pass

class PoolClosed(ThreadPoolException):
    pass

class ThreadPool:
    def __init__(self, size, paused=False):
        '''
        paused:
            The pool will start in a paused state and you will have to call
            `start` to start it.
        '''
        if not isinstance(size, int):
            raise TypeError(f'size must be an int, not {type(size)}.')
        if size < 1:
            raise ValueError(f'size must be >= 1, not {size}.')
        self.max_size = size
        self.closed = False
        self.paused = paused
        self._jobs = []
        self._job_manager_lock = threading.Lock()

    def _clear_done_jobs(self):
        '''
        This function assumes that _job_manager_lock is acquired!!
        You should call start instead!
        '''
        self._jobs[:] = [j for j in self._jobs if j.status in {PENDING, RUNNING}]

    def _start_jobs(self):
        '''
        This function assumes that _job_manager_lock is acquired!!
        You should call start instead!
        '''
        available = self.max_size - self.running_count
        available = max(0, available)
        if available == 0:
            return
        for job in list(self._jobs):
            if job.status == PENDING:
                job.start()
                available -= 1
                if available == 0:
                    break

    def _clear_done_and_start_jobs(self):
        '''
        This function assumes that _job_manager_lock is acquired!!
        You should call start instead!
        '''
        self._clear_done_jobs()
        self._start_jobs()

    def _job_finished(self):
        '''
        When a job finishes, it will call here.
        '''
        if self.paused:
            return

        # Although this method is private, we are calling the public `start`
        # instead of the private method because we do not hold the lock.
        self.start()

    @property
    def running_count(self):
        return sum(1 for job in list(self._jobs) if job.status is RUNNING)

    @property
    def unfinished_count(self):
        return sum(1 for job in list(self._jobs) if job.status in {PENDING, RUNNING})

    def assert_not_closed(self):
        '''
        If the pool is closed (because you called `join`), raise PoolClosed.
        Otherwise do nothing.
        '''
        if self.closed:
            raise PoolClosed()

    def add(self, function, *, name=None, args=tuple(), kwargs=dict()):
        '''
        Add a new job to the pool. Jobs are run in the order they are added.

        Don't forget that in order to write a tuple of length 1 you must still
        add a comma on the end. `add(print, args=(4))` is an error, you need to
        `add(print, args=(4,))` or use a list instead: `add(print, args=[4])`.

        name:
            An optional value that will appear in the repr of the job and
            has no other purpose. Use this if you intend to print(job) and want
            a human friendly name string.
        '''
        self.assert_not_closed()

        with self._job_manager_lock:
            job = Job(
                pool=self,
                function=function,
                name=name,
                args=args,
                kwargs=kwargs,
            )
            self._jobs.append(job)

            if not self.paused:
                self._clear_done_and_start_jobs()

        return job

    def add_many(self, kwargss):
        '''
        Add multiple new jobs to the pool at once. Useful to prevent the
        excessive lock-waiting that you get from calling regular `add` in a
        loop while other jobs are finishing and triggering queue maintenance.

        Provide an iterable of kwarg dictionaries. That is:
        [
            {'function': print, 'args': [4], 'name': '4'},
            {'function': sample, 'kwargs': {'x': 2}},
        ]
        '''
        self.assert_not_closed()

        with self._job_manager_lock:
            these_jobs = []
            for kwargs in kwargss:
                kwargs.pop('pool', None)
                job = Job(pool=self, **kwargs)
                these_jobs.append(job)
                self._jobs.append(job)

            if not self.paused:
                self._clear_done_and_start_jobs()

        return these_jobs

    def join(self):
        '''
        Permanently close the pool, preventing any new jobs from being added,
        and block until all jobs are complete.
        '''
        self.closed = True
        self.start()
        for job in self._jobs:
            job.join()

    def start(self):
        '''
        Remove finished and raised jobs from the queue and start some new jobs.

        The job queue is maintained automatically while adding new jobs and
        when a job finishes, as long as the pool is not paused, so you should
        not have to call it yourself. If you do pause the pool, use this method
        to restart it.

        Because the pool's internal job queue is flushed regularly, you should
        store your own references to jobs to get their return values.
        '''
        with self._job_manager_lock:
            self._clear_done_and_start_jobs()
            self.paused = False

class Job:
    def __init__(self, pool, function, *, name=None, args=tuple(), kwargs=dict()):
        self.pool = pool
        self.name = name
        self.status = PENDING
        self.function = function
        self.args = args
        self.kwargs = kwargs
        self.value = NO_RETURN
        self.exception = NO_EXCEPTION
        self._thread = None

        # _joinme_lock works because it is possible for a single thread to block
        # itself by calling `lock.acquire()` twice. The first call is here,
        # and the second call is in `join` so that join will block until the
        # lock is released by the job's finishing phase.
        self._joinme_lock = threading.Lock()
        self._joinme_lock.acquire()

    def __repr__(self):
        if self.name:
            return f'<{self.status} Job {repr(self.name)}>'
        else:
            return f'<{self.status} Job on {self.function}>'

    def join(self):
        '''
        Block until this job runs and completes.
        '''
        self._joinme_lock.acquire()
        self._joinme_lock.release()

    def start(self):
        '''
        Start the job. If the function completes successfully you will find the
        return value in `value`. If it raises an exception, you'll find it in
        `exception`, although the thread itself will not raise.
        '''
        def do_it():
            try:
                self.value = self.function(*self.args, **self.kwargs)
                self.status = FINISHED
            except Exception as exc:
                self.exception = exc
                self.status = RAISED
            self._thread = None
            self.pool._job_finished()
            self._joinme_lock.release()

        self.status = RUNNING
        self._thread = threading.Thread(target=do_it)
        self._thread.daemon = True
        self._thread.start()



YES_STRINGS = ['yes', 'y']
NO_STRINGS = ['no', 'n']

def getpermission(
        prompt=None,
        *,
        yes_strings=YES_STRINGS,
        no_strings=NO_STRINGS,
        must_pick=False,
    ):
    '''
    Prompt the user with a yes or no question.

    Return True for yes, False for no, and None if undecided.

    You can customize the strings that mean yes or no. For example, you could
    create a "type the name of the thing you're about to delete" prompt.

    If `must_pick`, then undecided is not allowed and the input will repeat
    until they choose an acceptable answer. Either way, the intended usage of
    `if getpermission():` will always only accept in case of explicit yes.
    '''
    if prompt is not None:
        print(prompt)
    while True:
        answer = input(f'{yes_strings[0]}/{no_strings[0]}> ').strip()
        yes = answer.lower() in (option.lower() for option in yes_strings)
        no = answer.lower() in (option.lower() for option in no_strings)
        if yes or no or not must_pick:
            break

    if yes:
        return True
    if no:
        return False
    return None



import hashlib
import logging
import os
import shutil
import sys

# pip install voussoirkit
from voussoirkit import bytestring
from voussoirkit import pathclass
from voussoirkit import ratelimiter

logging.basicConfig(level=logging.CRITICAL)
log = logging.getLogger(__name__)

# Number of bytes to read and write at a time
CHUNK_SIZE = 2 * bytestring.MIBIBYTE

HASH_CLASS = hashlib.md5

class SpinalException(Exception):
    pass

class DestinationIsDirectory(SpinalException):
    pass

class DestinationIsFile(SpinalException):
    pass

class RecursiveDirectory(SpinalException):
    pass

class SourceNotDirectory(SpinalException):
    pass

class SourceNotFile(SpinalException):
    pass

class SpinalError(SpinalException):
    pass

class ValidationError(SpinalException):
    pass

def callback_exclusion_v1(name, path_type):
    '''
    Example of an exclusion callback function.
    '''
    print('Excluding', path_type, name)

def callback_v1(fpobj, written_bytes, total_bytes):
    '''
    Example of a copy callback function.

    Prints "filename written/total (percent%)"
    '''
    filename = fpobj.absolute_path.encode('ascii', 'replace').decode()
    if written_bytes >= total_bytes:
        ends = '\r\n'
    else:
        ends = ''
    percent = (100 * written_bytes) / max(total_bytes, 1)
    percent = f'{percent:07.3f}'
    written = '{:,}'.format(written_bytes)
    total = '{:,}'.format(total_bytes)
    written = written.rjust(len(total), ' ')
    status = '{filename} {written}/{total} ({percent}%)\r'
    status = status.format(filename=filename, written=written, total=total, percent=percent)
    print(status, end=ends)
    sys.stdout.flush()

def copy(source, file_args=None, file_kwargs=None, dir_args=None, dir_kwargs=None):
    '''
    Perform copy_dir or copy_file as appropriate for the source path.
    '''
    source = pathclass.Path(source)
    if source.is_file:
        file_args = file_args or tuple()
        file_kwargs = file_kwargs or dict()
        return copy_file(source, *file_args, **file_kwargs)
    elif source.is_dir:
        dir_args = dir_args or tuple()
        dir_kwargs = dir_kwargs or dict()
        return copy_dir(source, *dir_args, **dir_kwargs)
    raise SpinalError(f'Neither file nor dir: {source}')

def copy_dir(
        source,
        destination=None,
        *,
        bytes_per_second=None,
        callback_directory=None,
        callback_exclusion=None,
        callback_file=None,
        callback_permission_denied=None,
        chunk_size=CHUNK_SIZE,
        destination_new_root=None,
        dry_run=False,
        exclude_directories=None,
        exclude_filenames=None,
        files_per_second=None,
        overwrite_old=True,
        precalcsize=False,
        stop_event=None,
        validate_hash=False,
    ):
    '''
    Copy all of the contents from source to destination,
    including subdirectories.

    source:
        The directory which will be copied.

    destination:
        The directory in which copied files are placed. Alternatively, use
        destination_new_root.

    bytes_per_second:
        Restrict file copying to this many bytes per second. Can be an integer
        or an existing Ratelimiter object.
        The BYTE, KIBIBYTE, etc constants from module 'bytestring' may help.

        Default = None

    callback_directory:
        This function will be called after each file copy with three parameters:
        name of file copied, number of bytes written to destination directory
        so far, total bytes needed (based on precalcsize).
        If `precalcsize` is False, this function will receive written bytes
        for both written and total, showing 100% always.

        Default = None

    callback_exclusion:
        Passed directly into `walk_generator`.

        Default = None

    callback_file:
        Will be passed into each individual `copy_file` operation as the
        `callback` for that file.

        Default = None

    callback_permission_denied:
        Will be passed into each individual `copy_file` operation as the
        `callback_permission_denied` for that file.

        Default = None

    destination_new_root:
        Determine the destination path by calling
        `new_root(source, destination_new_root)`.
        Thus, this path acts as a root and the rest of the path is matched.

        `destination` and `destination_new_root` are mutually exclusive.

    dry_run:
        Do everything except the actual file copying.

        Default = False

    exclude_filenames:
        Passed directly into `walk_generator`.

        Default = None

    exclude_directories:
        Passed directly into `walk_generator`.

        Default = None

    files_per_second:
        Maximum number of files to be processed per second. Helps to keep CPU
        usage low.

        Default = None

    overwrite_old:
        If True, overwrite the destination file if the source file
        has a more recent "last modified" timestamp.

        Default = True

    precalcsize:
        If True, calculate the size of source before beginning the
        operation. This number can be used in the callback_directory function.
        Else, callback_directory will receive written bytes as total bytes
        (showing 100% always).
        This can take a long time.

        Default = False

    stop_event:
        If provided, a threading.Event object which when set indicates that we
        should finish the current file and then stop the remainder of the copy.

        Default = None

    validate_hash:
        Passed directly into each `copy_file`.

    Returns: [destination path, number of bytes written to destination]
    (Written bytes is 0 if all files already existed.)
    '''
    # Prepare parameters
    if not is_xor(destination, destination_new_root):
        message = 'One and only one of `destination` and '
        message += '`destination_new_root` can be passed.'
        raise ValueError(message)

    source = pathclass.Path(source)
    source.correct_case()

    if destination_new_root is not None:
        destination = new_root(source, destination_new_root)

    destination = pathclass.Path(destination)
    destination.correct_case()

    if destination in source:
        raise RecursiveDirectory(source, destination)

    if not source.is_dir:
        raise SourceNotDirectory(source)

    if destination.is_file:
        raise DestinationIsFile(destination)

    if precalcsize is True:
        total_bytes = get_dir_size(source)
    else:
        total_bytes = 0

    callback_directory = callback_directory or do_nothing
    bytes_per_second = limiter_or_none(bytes_per_second)
    files_per_second = limiter_or_none(files_per_second)

    # Copy
    written_bytes = 0
    walker = walk_generator(
        source,
        callback_exclusion=callback_exclusion,
        exclude_directories=exclude_directories,
        exclude_filenames=exclude_filenames,
    )
    for source_file in walker:
        if stop_event and stop_event.is_set():
            break

        if source_file.is_link:
            continue

        destination_abspath = source_file.absolute_path.replace(
            source.absolute_path,
            destination.absolute_path
        )
        destination_file = pathclass.Path(destination_abspath)

        if destination_file.is_dir:
            raise DestinationIsDirectory(destination_file)

        if not dry_run:
            os.makedirs(destination_file.parent.absolute_path, exist_ok=True)

        copied = copy_file(
            source_file,
            destination_file,
            bytes_per_second=bytes_per_second,
            callback_progress=callback_file,
            callback_permission_denied=callback_permission_denied,
            chunk_size=chunk_size,
            dry_run=dry_run,
            overwrite_old=overwrite_old,
            validate_hash=validate_hash,
        )

        copiedname = copied[0]
        written_bytes += copied[1]

        if precalcsize is False:
            callback_directory(copiedname, written_bytes, written_bytes)
        else:
            callback_directory(copiedname, written_bytes, total_bytes)

        if files_per_second is not None:
            files_per_second.limit(1)

    return [destination, written_bytes]

def copy_file(
        source,
        destination=None,
        *,
        destination_new_root=None,
        bytes_per_second=None,
        callback_progress=None,
        callback_permission_denied=None,
        callback_validate_hash=None,
        chunk_size=CHUNK_SIZE,
        dry_run=False,
        overwrite_old=True,
        validate_hash=False,
    ):
    '''
    Copy a file from one place to another.

    source:
        The file to copy.

    destination:
        The filename of the new copy. Alternatively, use
        destination_new_root.

    destination_new_root:
        Determine the destination path by calling
        `new_root(source_dir, destination_new_root)`.
        Thus, this path acts as a root and the rest of the path is matched.

    bytes_per_second:
        Restrict file copying to this many bytes per second. Can be an integer
        or an existing Ratelimiter object.
        The provided BYTE, KIBIBYTE, etc constants may help.

        Default = None

    callback_permission_denied:
        If provided, this function will be called when a source file denies
        read access, with the file path and the exception object as parameters.
        THE OPERATION WILL RETURN NORMALLY.

        If not provided, the PermissionError is raised.

        Default = None

    callback_progress:
        If provided, this function will be called after writing
        each chunk_size bytes to destination with three parameters:
        the Path object being copied, number of bytes written so far,
        total number of bytes needed.

        Default = None

    callback_validate_hash:
        Passed directly into `verify_hash`

        Default = None

    dry_run:
        Do everything except the actual file copying.

        Default = False

    overwrite_old:
        If True, overwrite the destination file if the source file
        has a more recent "last modified" timestamp.

        Default = True

    validate_hash:
        If True, verify the file hash of the resulting file, using the
        `HASH_CLASS` global.

        Default = False

    Returns: [destination filename, number of bytes written to destination]
    (Written bytes is 0 if the file already existed.)
    '''
    # Prepare parameters
    if not is_xor(destination, destination_new_root):
        message = 'One and only one of `destination` and '
        message += '`destination_new_root` can be passed'
        raise ValueError(message)

    source = pathclass.Path(source)

    if not source.is_file:
        raise SourceNotFile(source)

    if destination_new_root is not None:
        source.correct_case()
        destination = new_root(source, destination_new_root)
    destination = pathclass.Path(destination)

    callback_progress = callback_progress or do_nothing

    if destination.is_dir:
        destination = destination.with_child(source.basename)

    bytes_per_second = limiter_or_none(bytes_per_second)

    # Determine overwrite
    if destination.exists:
        if overwrite_old is False:
            return [destination, 0]

        source_modtime = source.stat.st_mtime
        destination_modtime = destination.stat.st_mtime
        if source_modtime == destination_modtime:
            return [destination, 0]

    # Copy
    if dry_run:
        if callback_progress is not None:
            callback_progress(destination, 0, 0)
        return [destination, 0]

    source_bytes = source.size
    destination_location = os.path.split(destination.absolute_path)[0]
    os.makedirs(destination_location, exist_ok=True)

    def handlehelper(path, mode):
        try:
            handle = open(path.absolute_path, mode)
            return handle
        except PermissionError as exception:
            if callback_permission_denied is not None:
                callback_permission_denied(path, exception)
                return None
            else:
                raise

    log.debug('Opening handles.')
    source_handle = handlehelper(source, 'rb')
    destination_handle = handlehelper(destination, 'wb')

    if source_handle is None and destination_handle:
        destination_handle.close()
        return [destination, 0]

    if destination_handle is None:
        source_handle.close()
        return [destination, 0]

    if validate_hash:
        hasher = HASH_CLASS()

    written_bytes = 0
    while True:
        try:
            data_chunk = source_handle.read(chunk_size)
        except PermissionError as exception:
            if callback_permission_denied is not None:
                callback_permission_denied(source, exception)
                return [destination, 0]
            else:
                raise
        data_bytes = len(data_chunk)
        if data_bytes == 0:
            break

        if validate_hash:
            hasher.update(data_chunk)

        destination_handle.write(data_chunk)
        written_bytes += data_bytes

        if bytes_per_second is not None:
            bytes_per_second.limit(data_bytes)

        callback_progress(destination, written_bytes, source_bytes)

    if written_bytes == 0:
        # For zero-length files, we want to get at least one call in there.
        callback_progress(destination, written_bytes, source_bytes)

    # Fin
    log.debug('Closing source handle.')
    source_handle.close()
    log.debug('Closing dest handle.')
    destination_handle.close()
    log.debug('Copying metadata')
    shutil.copystat(source.absolute_path, destination.absolute_path)

    if validate_hash:
        verify_hash(
            destination,
            callback_progress=callback_validate_hash,
            known_size=source_bytes,
            known_hash=hasher.hexdigest(),
        )

    return [destination, written_bytes]

def do_nothing(*args):
    '''
    Used by other functions as the default callback.
    '''
    return

def get_dir_size(path):
    '''
    Calculate the total number of bytes across all files in this directory
    and its subdirectories.
    '''
    path = pathclass.Path(path)

    if not path.is_dir:
        raise SourceNotDirectory(path)

    total_bytes = 0
    for filepath in walk_generator(path):
        total_bytes += filepath.size

    return total_bytes

def hash_file(
        path,
        hash_class=HASH_CLASS,
        *,
        callback_progress=do_nothing,
        chunk_size=CHUNK_SIZE,
    ):
    '''
    callback_progress:
        A function that takes three parameters:
        path object, bytes ingested so far, bytes total
    '''
    path = pathclass.Path(path)
    hasher = hash_class()
    checked_bytes = 0
    file_size = os.path.getsize(path.absolute_path)

    handle = open(path.absolute_path, 'rb')
    with handle:
        while True:
            chunk = handle.read(chunk_size)
            if not chunk:
                break
            hasher.update(chunk)
            checked_bytes += len(chunk)
            callback_progress(path, checked_bytes, file_size)

    return hasher.hexdigest()

def is_xor(*args):
    '''
    Return True if and only if one arg is truthy.
    '''
    return [bool(a) for a in args].count(True) == 1

def limiter_or_none(value):
    if isinstance(value, str):
        value = bytestring.parsebytes(value)
    if isinstance(value, ratelimiter.Ratelimiter):
        limiter = value
    elif value is not None:
        limiter = ratelimiter.Ratelimiter(allowance=value, period=1)
    else:
        limiter = None
    return limiter

def new_root(filepath, root):
    '''
    Prepend `root` to `filepath`, drive letter included. For example:
    "C:\\folder\\subfolder\\file.txt" and "C:\\backups" becomes
    "C:\\backups\\C\\folder\\subfolder\\file.txt"

    I use this so that my G: drive can have backups from my C: and D: drives
    while preserving directory structure in G:\\D and G:\\C.
    '''
    filepath = pathclass.Path(filepath).absolute_path
    root = pathclass.Path(root).absolute_path
    filepath = filepath.replace(':', os.sep)
    filepath = os.path.normpath(filepath)
    filepath = os.path.join(root, filepath)
    return pathclass.Path(filepath)

def normalize(text):
    '''
    Apply os.path.normpath and os.path.normcase.
    '''
    return os.path.normpath(os.path.normcase(text))

def verify_hash(
        path,
        known_hash,
        *,
        known_size=None,
        **hash_kwargs,
    ):
    path = pathclass.Path(path)
    log.debug('Validating hash for "%s" against %s', path.absolute_path, known_hash)

    if known_size is not None:
        file_size = os.path.getsize(path.absolute_path)
        if file_size != known_size:
            raise ValidationError(f'File size {file_size} != known size {known_size}.')

    file_hash = hash_file(path, **hash_kwargs)
    if file_hash != known_hash:
        raise ValidationError(f'File hash "{file_hash}" != known hash "{known_hash}".')
    log.debug('Hash validation passed.')

def walk_generator(
        path='.',
        *,
        callback_exclusion=None,
        callback_permission_denied=None,
        exclude_directories=None,
        exclude_filenames=None,
        recurse=True,
        yield_directories=False,
        yield_files=True,
        yield_style='flat',
    ):
    '''
    Yield Path objects for files in the file tree, similar to os.walk.

    callback_exclusion:
        This function will be called when a file or directory is excluded with
        two parameters: the path, and 'file' or 'directory'.

        Default = None

    exclude_filenames:
        A set of filenames that will not be copied. Entries can be absolute
        paths to exclude that particular file, or plain names to exclude
        all matches. For example:
        {'C:\\folder\\file.txt', 'desktop.ini'}

        Default = None

    exclude_directories:
        A set of directories that will not be copied. Entries can be
        absolute paths to exclude that particular directory, or plain names
        to exclude all matches. For example:
        {'C:\\folder', 'thumbnails'}

        Default = None

    recurse:
        Yield from subdirectories. If False, only immediate files are returned.

    yield_directories:
        Should the generator produce directories? Has no effect in nested yield style.

    yield_files:
        Should the generator produce files? Has no effect in nested yield style.

    yield_style:
        If 'flat', yield individual files one by one in a constant stream.
        If 'nested', yield tuple(root, directories, files) like os.walk does,
            except I use Path objects with absolute paths for everything.
    '''
    if not yield_directories and not yield_files:
        raise ValueError('yield_directories and yield_files cannot both be False.')

    if yield_style not in ['flat', 'nested']:
        raise ValueError(f'yield_style should be "flat" or "nested", not {yield_style}.')

    if exclude_directories is None:
        exclude_directories = set()

    if exclude_filenames is None:
        exclude_filenames = set()

    callback_exclusion = callback_exclusion or do_nothing
    callback_permission_denied = callback_permission_denied or do_nothing
    _callback_permission_denied = callback_permission_denied
    def callback_permission_denied(error):
        return _callback_permission_denied(error.filename, error)

    exclude_filenames = {normalize(f) for f in exclude_filenames}
    exclude_directories = {normalize(f) for f in exclude_directories}

    path = pathclass.Path(path)
    path.correct_case()

    exclude = (
        path.basename in exclude_directories or
        path.absolute_path in exclude_directories
    )
    if exclude:
        callback_exclusion(path, 'directory')
        return

    def handle_exclusion(blacklist, basename, abspath, kind):
        exclude = (
            os.path.normcase(basename) in blacklist or
            os.path.normcase(abspath) in blacklist
        )
        if exclude:
            callback_exclusion(abspath, kind)
            return 1

    # In the following loops, I found joining the os.sep with fstrings to be
    # 10x faster than `os.path.join`, reducing a 6.75 second walk to 5.7.
    # Because we trust the values of current_location and the child names,
    # we don't run the risk of producing bad values this way.

    def walkstep_nested(current_location, child_dirs, child_files):
        directories = []
        new_child_dirs = []
        for child_dir in child_dirs:
            child_dir_abspath = f'{current_location}{os.sep}{child_dir}'
            if handle_exclusion(exclude_directories, child_dir, child_dir_abspath, 'directory'):
                continue

            new_child_dirs.append(child_dir)
            directories.append(pathclass.Path(child_dir_abspath))

        # This will actually affect the results of the os.walk going forward!
        child_dirs[:] = new_child_dirs

        files = []
        for child_file in child_files:
            child_file_abspath = f'{current_location}{os.sep}{child_file}'
            if handle_exclusion(exclude_filenames, child_file, child_file_abspath, 'file'):
                continue

            files.append(pathclass.Path(child_file_abspath))

        current_location = pathclass.Path(current_location)
        yield (current_location, directories, files)

    def walkstep_flat(current_location, child_dirs, child_files):
        new_child_dirs = []
        for child_dir in child_dirs:
            child_dir_abspath = f'{current_location}{os.sep}{child_dir}'
            if handle_exclusion(exclude_directories, child_dir, child_dir_abspath, 'directory'):
                continue

            new_child_dirs.append(child_dir)
            if yield_directories:
                yield pathclass.Path(child_dir_abspath)

        # This will actually affect the results of the os.walk going forward!
        child_dirs[:] = new_child_dirs

        if yield_files:
            for child_file in child_files:
                child_file_abspath = f'{current_location}{os.sep}{child_file}'
                if handle_exclusion(exclude_filenames, child_file, child_file_abspath, 'file'):
                    continue

                yield pathclass.Path(child_file_abspath)

    walker = os.walk(path.absolute_path, onerror=callback_permission_denied)
    if yield_style == 'flat':
        for step in walker:
            yield from walkstep_flat(*step)
            if not recurse:
                break

    if yield_style == 'nested':
        for step in walker:
            yield from walkstep_nested(*step)
            if not recurse:
                break



import setuptools

setuptools.setup(
    name='voussoirkit',
    packages=setuptools.find_packages(),
    version='0.0.50',
    author='voussoir',
    author_email='pypi@voussoir.net',
    description='voussoir\'s toolkit',
    long_description=open('README.md', 'r').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/voussoir/voussoirkit',
    install_requires=[
        'pyperclip',
        'pywin32;platform_system=="Windows"',
        'winshell;platform_system=="Windows"',
    ]
)



'''
NOTE: I have not yet figured out how to get a list of volumes that are mounted
to folders instead of drive letters. I think this is the API I need:
https://docs.microsoft.com/en-us/windows/win32/fileio/enumerating-volume-mount-points
'''
import ctypes
import string

def get_drive_info(letter):
    '''
    Given a drive letter, return a dictionary containing its attributes.

    Thanks Nicholas Orlowski
    http://stackoverflow.com/a/12056414
    '''
    kernel32 = ctypes.windll.kernel32
    name_buffer = ctypes.create_unicode_buffer(1024)
    filesystem_buffer = ctypes.create_unicode_buffer(1024)
    serial_number = None
    max_component_length = None
    file_system_flags = None

    letter = letter.rstrip(':\\/')
    drive_active = kernel32.GetVolumeInformationW(
        ctypes.c_wchar_p(f'{letter}:\\'),
        name_buffer,
        ctypes.sizeof(name_buffer),
        serial_number,
        max_component_length,
        file_system_flags,
        filesystem_buffer,
        ctypes.sizeof(filesystem_buffer),
    )
    info = {
        'active': bool(drive_active),
        'filesystem': filesystem_buffer.value,
        'name': name_buffer.value,
    }
    return info

def get_drive_letters():
    '''
    Return a list of all connected drive letters.

    Thanks RichieHindle
    http://stackoverflow.com/a/827398

    "If the function succeeds, the return value is a bitmask representing the
    currently available disk drives. Bit position 0 (the least-significant bit)
    is drive A, bit position 1 is drive B, bit position 2 is drive C, and so on."
    https://docs.microsoft.com/en-us/windows/win32/api/fileapi/nf-fileapi-getlogicaldrives
    '''
    letters = []
    bitmask = ctypes.windll.kernel32.GetLogicalDrives()
    for letter in string.ascii_uppercase:
        if bitmask & 1:
            letters.append(letter)
        bitmask >>= 1
    return letters

def get_drive_letter_by_name(name):
    '''
    Given the name of a drive (the user-customizable name seen in Explorer),
    return the letter of that drive.

    Raises KeyError if it is not found.
    '''
    drives = get_drive_map()
    for (letter, info) in drives.items():
        if info['name'] == name:
            return letter
    raise KeyError(name)

def get_drive_map():
    '''
    Return a dict of {letter: info}.
    '''
    drives = {letter: get_drive_info(letter) for letter in get_drive_letters()}
    return drives



'''
This module is designed to provide a GOOD ENOUGH means of identifying duplicate
files very quickly, so that more in-depth checks can be done on likely matches.
'''

import hashlib
import os
import sys

SEEK_END = 2
CHUNK_SIZE = 2**20
FORMAT = '{size}_{hashtype}_{chunk_size}_{hash}'

def equal_handle(handle1, handle2, *args, **kwargs):
    size1 = handle1.seek(0, SEEK_END)
    size2 = handle2.seek(0, SEEK_END)
    handle1.seek(0)
    handle2.seek(0)
    if size1 != size2:
        return False
    id1 = quickid_handle(handle1, *args, **kwargs)
    id2 = quickid_handle(handle2, *args, **kwargs)
    return id1 == id2

def equal_file(filename1, filename2, *args, **kwargs):
    filename1 = os.path.abspath(filename1)
    filename2 = os.path.abspath(filename2)
    if os.path.getsize(filename1) != os.path.getsize(filename2):
        return False
    with open(filename1, 'rb') as handle1, open(filename2, 'rb') as handle2:
        return equal_handle(handle1, handle2, *args, **kwargs)

def quickid_handle(handle, chunk_size=None):
    if chunk_size is None:
        chunk_size = CHUNK_SIZE

    hashtype = 'md5'
    hasher = hashlib.md5()
    size = handle.seek(0, SEEK_END)
    handle.seek(0)

    if size <= 2 * chunk_size:
        hasher.update(handle.read())
    else:
        hasher.update(handle.read(chunk_size))
        handle.seek(-1 * chunk_size, SEEK_END)
        hasher.update(handle.read())

    output = FORMAT.format(
        size=size,
        hashtype=hashtype,
        chunk_size=chunk_size,
        hash=hasher.hexdigest(),
    )
    return output

def quickid_file(filename, *args, **kwargs):
    filename = os.path.abspath(filename)
    with open(filename, 'rb') as handle:
        return quickid_handle(handle, *args, **kwargs)

def main(argv):
    print(quickid_file(argv[0]))

if __name__ == '__main__':
    raise SystemExit(main(sys.argv[1:]))
